#include <asm/msr-index.h>

#include "shared-cpuid.c"

static void emulate_cpuid(struct regs *regs)
{
    struct kvm_cpuid_entry entry;

    entry.function = regs->rax;
    real_cpuid(&entry);
    filter_cpuid(&entry, longmode);
    regs->rax = entry.eax;
    regs->rbx = entry.ebx;
    regs->rcx = entry.ecx;
    regs->rdx = entry.edx;
}

static void emulate_rdmsr(struct regs *regs)
{
    uint32_t ax,dx;
    rdmsr(regs->rcx, &ax, &dx);
    regs->rax = ax;
    regs->rdx = dx;
}

static void emulate_wrmsr(struct regs *regs)
{
    static const uint64_t known = (EFER_NX|EFER_LMA|EFER_LME|EFER_SCE);
    static const uint64_t fixed = (EFER_LMA|EFER_LME|EFER_SCE);
    uint32_t ax,dx;
    
    switch (regs->rcx) {
    case MSR_EFER:
	if (regs->rax & ~known) {
	    printk("%s: efer: unknown bit set\n", __FUNCTION__);
	    goto out;
	}

	rdmsr(regs->rcx, &ax, &dx);
	if ((regs->rax & fixed) != (ax & fixed)) {
	    printk("%s: efer: modify fixed bit\n", __FUNCTION__);
	    goto out;
	}

	printk("%s: efer:%s%s%s%s\n", __FUNCTION__,
	       regs->rax & EFER_SCE ? " sce" : "",
	       regs->rax & EFER_LME ? " lme" : "",
	       regs->rax & EFER_LMA ? " lma" : "",
	       regs->rax & EFER_NX  ? " nx"  : "");
	/* fall through */
    case MSR_FS_BASE:
    case MSR_GS_BASE:
    case MSR_KERNEL_GS_BASE:
	wrmsr(regs->rcx, regs->rax, regs->rdx);
	return;
    }

out:
    printk("%s: ignore: 0x%" PRIxREG " 0x%" PRIxREG ":0x%" PRIxREG "\n",
	   __FUNCTION__, regs->rcx, regs->rdx, regs->rax);
}

static void print_emu_instr(char *prefix, uint8_t *instr)
{
    printk("%s: rip %p bytes %02x %02x %02x %02x  %02x %02x %02x %02x\n",
	   prefix, instr,
	   instr[0], instr[1], instr[2], instr[3],
	   instr[4], instr[5], instr[6], instr[7]);
}

static ureg_t *decode_reg(struct regs *regs, uint8_t modrm, int rm)
{
    int shift = rm ? 0 : 3;
    ureg_t *reg = NULL;

    switch ((modrm >> shift) & 0x07) {
    case 0: reg = (ureg_t*)&regs->rax;
    case 1: reg = (ureg_t*)&regs->rcx;
    case 2: reg = (ureg_t*)&regs->rdx;
    case 3: reg = (ureg_t*)&regs->rbx;
    case 4: reg = (ureg_t*)&regs->rsp;
    case 5: reg = (ureg_t*)&regs->rbp;
    case 6: reg = (ureg_t*)&regs->rsi;
    case 7: reg = (ureg_t*)&regs->rdi;
    }
    return reg;
}

int emulate(struct regs *regs)
{
    static const uint8_t xen_emu_prefix[5] = {0x0f, 0x0b, 'x','e','n'};
    uint8_t *instr;
    int skip = 0;
    int in = 0;
    int shift = 0;

restart:
    instr = (void*)regs->rip;

    /* prefixes */
    if (instr[skip] == 0x66) {
	shift = 16;
	skip++;
    }

    /* instructions */
    switch (instr[skip]) {
    case 0x0f:
	switch (instr[skip+1]) {
	case 0x09:
	    /* wbinvd -- ignore for now, vmexit should come close enougth ;) */
	    __asm__("wbinvd" ::: "memory");
	    skip += 2;
	    break;
	case 0x0b:
	    /* ud2a */
	    if (xen_emu_prefix[2] == instr[skip+2] &&
		xen_emu_prefix[3] == instr[skip+3] &&
		xen_emu_prefix[4] == instr[skip+4]) {
		printk("%s: xen emu prefix\n", __FUNCTION__);
		regs->rip += 5;
		goto restart;
	    }
	    return 0; /* bounce to kernel so we get a fancy trace */
	case 0x20:
	{
	    /* read control registers */
	    ureg_t *reg = decode_reg(regs, instr[skip+2], 1);
	    switch (((instr[skip+2]) >> 3) & 0x07) {
	    case 0:
		*reg = read_cr0();
		skip = 3;
		break;
	    case 3:
		*reg = read_cr3();
		skip = 3;
		break;
	    case 4:
		*reg = read_cr4();
		skip = 3;
		break;
	    }
	    break;
	}
	case 0x22:
	{
	    /* write control registers */
	    ureg_t *reg = decode_reg(regs, instr[skip+2], 1);
	    ureg_t cr;
	    switch (((instr[skip+2]) >> 3) & 0x07) {
	    case 0:
		cr = read_cr0();
		if (cr != *reg)
		    printk("%s: ignore cr0 write (0x%" PRIxREG " -> 0x%" PRIxREG ")\n",
			   __FUNCTION__, cr, *reg);
		skip = 3;
		break;
	    case 4:
		cr = read_cr4();
		if (cr != *reg)
		    printk("%s: ignore cr4 write (0x%" PRIxREG " -> 0x%" PRIxREG ")\n",
			   __FUNCTION__, cr, *reg);
		skip = 3;
		break;
	    }
	    break;
	}
	case 0x30:
	    /* wrmsr */
	    emulate_wrmsr(regs);
	    skip += 2;
	    break;
	case 0x32:
	    /* rdmsr */
	    emulate_rdmsr(regs);
	    skip += 2;
	    break;
	case 0xa2:
	    /* cpuid */
	    emulate_cpuid(regs);
	    skip += 2;
	    break;
	}
	break;

    case 0xe4: /* in     <next byte>,%al */
    case 0xe5:
	skip += 2;
	in = (instr[0] & 1) ? 2 : 1;
	break;
    case 0xec: /* in     (%dx),%al */
    case 0xed:
	skip += 1;
	in = (instr[0] & 1) ? 2 : 1;
	break;
    case 0xe6: /* out    %al,<next byte> */
    case 0xe7:
	skip += 2;
	break;
    case 0xee: /* out    %al,(%dx) */
    case 0xef:
	skip += 1;
	break;

    case 0xfa:
	/* cli */
	shared_info.vcpu_info[0].evtchn_upcall_mask = 1;
	skip += 1;
	break;
    }

    /* unknown instruction */
    if (!skip) {
	print_emu_instr("instr emu failed", instr);
	return -1;
    }

    /* I/O instruction */
    if (2 == in)
	regs->rax |= 0xffffffff;
    if (1 == in)
	regs->rax |= (0xffff << shift);

    return skip;
}
